# include <vector>
# include <iostream>

using namespace std;

# include "IOUtils.h"
# include "MathUtils.h"
# include "MinMax.h"
# include "CalcROC.h"

Real calcFPR(Vector &x, Vector &y, Real &xCutoff, Real &yCutoff)
{
	int Size1=x.size();
	int Size2=y.size();
	int negative=0, nfp=0;
	Real fpr;

	if (Size1!=Size2 || Size1==0)
	{
		string errorStr="Size1= "+toStr(Size1)+" Size2= "+toStr(Size2);
		error(errorStr, __LINE__, __FILE__);
	}
	//cout <<"xCutoff= "<<xCutoff<<endl;
	for (int i=0;i<Size1;i++)
	{
		if (y[i]>yCutoff) negative++;
		if (y[i]>yCutoff && x[i]<xCutoff) 
		{
			nfp++;
			//cout <<x[i]<<"\t"<<y[i]<<"\tFalse positive"<<endl;
		}
		else
		{
			//cout <<x[i]<<"\t"<<y[i]<<endl;
		}
	}
	if (negative>0) fpr=Real(nfp)/Real(negative);
	else fpr=0;
	//cout <<"xCutoff= "<<xCutoff<<" negative= "<<negative<<" nfp= "<<nfp<<" fpr= "<<fpr<<endl;
	return fpr;
}

Real calcFNR(Vector &x, Vector &y, Real &xCutoff, Real &yCutoff)
{
	int Size1=x.size();
	int Size2=y.size();
	int positive=0, nfn=0;
	Real fpr;

	if (Size1!=Size2 || Size1==0)
	{
		string errorStr="Size1= "+toStr(Size1)+" Size2= "+toStr(Size2);
		error(errorStr, __LINE__, __FILE__);
	}

	for (int i=0;i<Size1;i++)
	{
		if (y[i]<yCutoff) positive++;
		if (y[i]<yCutoff && x[i]>xCutoff) nfn++;
	}
	if (positive>0) fpr=Real(nfn)/Real(positive);
	else fpr=0;
	return fpr;
}

Real calcTPR(Vector &x, Vector &y, Real &xCutoff, Real &yCutoff)
{
	int Size1=x.size();
	int Size2=y.size();
	int positive=0, ntp=0;
	Real tpr;

	if (Size1!=Size2 || Size1==0)
	{
		string errorStr="Size1= "+toStr(Size1)+" Size2= "+toStr(Size2);
		error(errorStr, __LINE__, __FILE__);
	}

	for (int i=0;i<Size1;i++)
	{
		if (y[i]<yCutoff) positive++;
		if (y[i]<yCutoff && x[i]<xCutoff) ntp++;
	}
	if (positive>0) tpr=Real(ntp)/Real(positive);
	else tpr=0;
	//cout <<"xCutoff= "<<xCutoff<<" positive= "<<positive<<" ntp= "<<ntp<<" tpr= "<<tpr<<endl;
	return tpr;
}

Real calcROC(Vector &x, Vector &y, Real &yCutoff, vector<Real> &fprs, vector<Real> &tprs)
{
	int nbin=1000;
	Real xCutoff;
	Real xmin, xmax, xinc;
	Real area, fpr, tpr;

	fprs.clear();
	tprs.clear();

	xmin=findMin(x);
	xmax=findMax(x);;
	xinc=(xmax-xmin)/nbin;

	SafePushBack(fprs, Real(0.0), "fprs");
	SafePushBack(tprs, Real(0.0), "tprs");
	for (xCutoff=xmin;xCutoff<=xmax;xCutoff+=xinc)
	{
		fpr=calcFPR(x, y, xCutoff, yCutoff);
		tpr=calcTPR(x, y, xCutoff, yCutoff);
//		cout <<fpr<<"\t"<<tpr<<"\t"<<fnr<<endl;
		SafePushBack(fprs, fpr, "fprs");
		SafePushBack(tprs, tpr, "tprs");
	}
	SafePushBack(fprs, Real(1.0), "fprs");
	SafePushBack(tprs, Real(1.0), "tprs");
	area=calcAreaUnderCurve(fprs, tprs);

	printROC(fprs, tprs);

	cout <<"area= "<<area<<endl;

	return area;
}

Real calcROC(Vector &x, Vector &y, Real &yCutoff)
{
	Vector fprs, tprs;

	return calcROC(x, y, yCutoff, fprs, tprs);
}

void printROC(vector<Real> &fprs, vector<Real> &tprs)
{
	int Size1=fprs.size();
	int Size2=tprs.size();

	if (Size1!=Size2)
	{
		cout <<"ERROR: Cannot print ROC"<<endl;
	}

	cout <<"fpr\ttpr"<<endl;
	for (int i=0;i<Size1;i++)
	{
		if (fprs[i]!=0 && tprs[i]!=0)
		{
			cout <<fprs[i]<<"\t"<<tprs[i]<<endl;
		}
	}
}

