/*
    JV Library
    Copyright (C) 2014  Jouko Juhani Virtanen

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
# include <vector>
# include <iostream>

using namespace std;

# include "TypeDef.h"
# include "Structures.h"
# include "AtomUtils.h"
# include "PrintPdb.h"

Real LJPotential(Real epsilon, Real VDWRadius, Real r)
{
	Real vr, vr2, vr6;
	vr=VDWRadius/r;
	vr2=vr*vr;
	vr6=vr2*vr2*vr2;
	return 4.0*epsilon*(vr6*vr6-2.0*vr6);
}

inline Real LJPotential2(Real epsilon, Real VDWRadius, Real r2)
{
	Real vr2, vr6;
	Real energy;
	vr2=VDWRadius*VDWRadius/r2;
	vr6=vr2*vr2*vr2;
	energy=4.0*epsilon*(vr6*vr6-2.0*vr6);
	return energy;
}

Real LJPotential(AtomStruct &Atom1, AtomStruct &Atom2)
{
        Real r2=atomSquareDistance(Atom1, Atom2);
	if (r2>36.0) return 0;
        Real epsilon12=sqrt(Atom1.epsilon*Atom2.epsilon);
        Real VDWRadius12=Atom1.vdw+Atom2.vdw;
	return LJPotential2(epsilon12, VDWRadius12, r2);
}

Real LJPotential(vector<AtomStruct> &Atoms)
{
	int natom=Atoms.size();
	Real energy=0;

	for (int i=0;i<natom;i++)
	{
		for (int j=i+1;j<natom;j++)
		{
			if (!isBonded(Atoms[i], j))
			{
				energy+=LJPotential(Atoms[i], Atoms[j]);
			}
		}
	}
	return energy;
}

Real LJPotential(vector<AtomStruct> &Atoms1, vector<AtomStruct> &Atoms2)
{
	int natom1=Atoms1.size();
	int natom2=Atoms2.size();
	Real energy=0;
	for (int i=0;i<natom1;i++)
	{
		for (int j=0;j<natom2;j++)
		{
			energy+=LJPotential(Atoms1[i], Atoms2[j]);
		}
	}
	return energy;
}

//void makeLookUpTable(Real epsilon, Real VDWRadius, LJLookUpStruct &lookUp)
//{
//
//}

Real InvertLJPotential(Real epsilon, Real VDWRadius, Real energy)
{
	Real r, r6, sq;
	sq=4.0+energy/epsilon;
	if (sq<0)
	{
		cout <<"Error there is no distance that corresponds to that LJ potential"<<endl;
		return 0;
	}
	r6=2.0/(2.0+sqrt(sq));
	r=VDWRadius*exp(log(r6)/6.0);
	return r;
}
