/*
    JV Library
    Copyright (C) 2014  Jouko Juhani Virtanen

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
# include <vector>
# include <iostream>

using namespace std;

# include "TypeDef.h"
# include "Structures.h"
# include "MathUtils.h"
# include "VectorManip.h"
# include "AtomUtils.h"
# include "test_functions.h"
# include "minimize.h"

Real calcRestraintEnergy(ProteinStruct &Protein, vector<AtomStruct> &restraintAtoms, vector<int> &restraintAtomsIndex)
{
        Real sum=0;

        for (int i=0;i<2;i++)
        {
                sum+=atomSquareDistance(Protein.Atoms[restraintAtomsIndex[i]], restraintAtoms[i]);
        }
        return sum;
}

Real calcRestraintEnergy(vector<Real> &dihedral, ProteinStruct &Protein, vector<AtomStruct> &restraintAtoms, vector<int> &restraintAtomsIndex, vector<int> &variableDihedrals)
{
        Real sum=0;
        for (int i=0;i<int(variableDihedrals.size());i++)
        {
                setDihedralAngle(Protein.Atoms, variableDihedrals[i], dihedral[i]);
        }
        sum=calcRestraintEnergy(Protein, restraintAtoms, restraintAtomsIndex);
        return sum;
}

void setVariableDihedrals(vector<Real> &dihedrals, vector<int> &variableDihedrals, vector<Real> &protDihedrals)
{
        int ndihedral=variableDihedrals.size();

        SafeAlloc(dihedrals, ndihedral, "dihedrals in setVariableDihedrals");
        for (int i=0;i<ndihedral;i++)
        {
                dihedrals[i]=protDihedrals[variableDihedrals[i]];
        }
}

void getRestraintAtoms(vector<AtomStruct> &Atoms, vector<AtomStruct> &restraintAtoms, vector<int> &restraintAtomsIndex, int residue, int angleType)
{
        restraintAtomsIndex.resize(3);
        restraintAtomsIndex[0]=getAtomIndex(Atoms, residue+3, "CA");
        restraintAtomsIndex[1]=getAtomIndex(Atoms, residue+3, "C");
        restraintAtomsIndex[2]=getAtomIndex(Atoms, residue+4, "N");
        for (int i=0;i<3;i++)
        {
                SafePushBack(restraintAtoms, Atoms[restraintAtomsIndex[i]], "restraintAtoms");
                //cout <<"restraintAtomsIndex["<<i<<"]= "<<restraintAtomsIndex[i]<<endl;
        } 
}

vector<int> getVariableDihedrals(int residue)
{
        vector<int> variableDihedrals;
        //variableDihedrals.resize(6);
        variableDihedrals.resize(5);
        variableDihedrals[0]=residue*3-2;
        variableDihedrals[1]=residue*3-1;
        variableDihedrals[2]=residue*3+1;
        variableDihedrals[3]=residue*3+2;
        variableDihedrals[4]=residue*3+4;
        //variableDihedrals[5]=residue*3+5;
        return variableDihedrals;
}
/*
Real localMove(ProteinStruct &Protein, int residue, int angleType, Real dAngle)
{
        int numMoves;
        vector<int> restraintAtomsIndex, variableDihedrals;
        Real ddAngle, maxDAngle=1.0*DEGREE_TO_RAD;
        vector<AtomStruct> restraintAtoms;
        //cout <<"About to getRestraints"<<endl;
        getRestraintAtoms(Protein.Atoms, restraintAtoms, restraintAtomsIndex, residue, angleType);
        //cout <<"About to getVariableDihedrals"<<endl;
        variableDihedrals=getVariableDihedrals(residue);
        numMoves=int(dAngle/maxDAngle+0.5);
        ddAngle=dAngle/Real(numMoves);
        for (int i=0;i<numMoves;i++)
        {
                //cout <<"About to pivot"<<endl;
                pivot(Protein.Atoms, residue, angleType, ddAngle);
                //cout <<"About to minimizeRestraints"<<endl;
                minimizeRestraints(Protein, restraintAtoms, restraintAtomsIndex, variableDihedrals);
        }
        //cout <<"Leaving localMove"<<endl;
}
*/
void getBondAtoms(vector<AtomStruct> &Atoms, int index, int &atom1, int &atom2)
{
        int angleType=PHI, atom0, atom3;
        int residue=int((index+1)/3)+1;
        if (index%3==0) angleType=PSI;
        else if (index%3==1) angleType=OMEGA;
        else if (index%3==2) angleType=PHI;
        getDihedralAtoms(Atoms, atom0, atom1, atom2, atom3, angleType, residue);
}

void allignBondWithXAxis(ProteinStruct &Protein, vector<AtomStruct> &restraintAtoms, int index)
{
        bool boolPrint=false;
        int atom1, atom2;
        Real phi, dx, dy, dz;
        VectorStruct v1, vx, cp;
        //cout <<endl;
        getBondAtoms(Protein.Atoms, index, atom1, atom2);
        //printAtomInfo(Protein.Atoms[atom1]);
        //printAtomInfo(Protein.Atoms[atom2]);
        //cout <<"atom1= "<<atom1<<" atom2= "<<atom2<<endl;
        dx=-Protein.Atoms[atom2].x;
        dy=-Protein.Atoms[atom2].y;
        dz=-Protein.Atoms[atom2].z;
        moveAtoms(Protein.Atoms, dx, dy, dz, boolPrint);
        v1=atomVector(Protein.Atoms[atom2], Protein.Atoms[atom1]);
        //cout <<"v1= "<<vectorToStr(v1)<<endl;
        vx.x=1.0; vx.y=0.0; vx.z=0.0;
        cp=calcCrossProduct(vx, v1);
        phi=calcAngle(v1, vx);
        vectorNormalize(cp);
        rotateAtomsAroundVector(Protein.Atoms, cp, 0.0, 0.0, 0.0, -phi);
        //printAtomInfo(Protein.Atoms[atom1]);
        //printAtomInfo(Protein.Atoms[atom2]);
        //phi=calcVectorAngle(v1, vx);
        moveAtoms(restraintAtoms, dx, dy, dz, boolPrint);
        rotateAtomsAroundVector(restraintAtoms, cp, 0.0, 0.0, 0.0, -phi);
        //cout <<endl;
}

Real calcRestraintDerivative(AtomStruct &Atom1, AtomStruct &Atom2)
{
        //cout <<endl;
        //printAtomInfo(Atom1);
        //printAtomInfo(Atom2);
        //cout <<endl;
        return 2.0*Atom2.y*Atom1.z-2.0*Atom1.y*Atom2.z;
}

Real calcRestraintDerivative(AtomStruct &mobileAtom, AtomStruct &restraintAtom, AtomStruct &bondAtom, VectorStruct &bond)
{
        Real sum=0;
        Real dx, dy, dz;

        dx=restraintAtom.x-bondAtom.x;
        dy=restraintAtom.y-bondAtom.y;
        dz=restraintAtom.z-bondAtom.z;
        sum=(mobileAtom.x-restraintAtom.x)*(bond.z*dy-bond.y*dz);
        sum+=(mobileAtom.y-restraintAtom.y)*(bond.x*dz-bond.z*dx);
        sum+=(mobileAtom.z-restraintAtom.z)*(bond.y*dx-bond.x*dy);
        sum*=2.0;
        return sum;
}

Real calcRestraintDerivative(ProteinStruct &Protein, vector<AtomStruct> &restraintAtoms, vector<int> &restraintAtomsIndex, int &variableDihedral, int &atom1, VectorStruct &bond, int &index)
{
        //int nrestraint=restraintAtoms.size();
        Real sum=0;
        //allignBondWithXAxis(Protein, restraintAtoms, variableDihedral);
        vectorNormalize(bond);
        for (int i=0;i<2;i++)
        {
                sum+=calcRestraintDerivative(Protein.Atoms[restraintAtomsIndex[i]], restraintAtoms[i], Protein.Atoms[atom1], bond);
        }
        return sum;
}
/*
vector<Real> calcLocalMoveGradient(ProteinStruct &Protein, vector<AtomStruct> &restraintAtoms, vector<int> &restraintAtomsIndex, vector<int> &variableDihedrals)
{
        int ndihedral=variableDihedrals.size();
        vector<Real> gradient;
        SafeAlloc(gradient, ndihedral, "gradient");
        for (int i=0;i<ndihedral;i++)
        {
                gradient[i]=calcRestraintDerivative(Protein, restraintAtoms, restraintAtomsIndex, variableDihedrals[i], i);
                //cout <<"gradient["<<i<<"]= "<<gradient[i]<<endl;
        }
        return gradient;
}
*/
Real calcRestraintHessian(AtomStruct &Atom1, AtomStruct &Atom2, AtomStruct &Atom3, AtomStruct &Atom4)
{
        Real hessian;
        VectorStruct v;

        v=atomVector(Atom3, Atom2);
        vectorNormalize(v);
        hessian=2.0*Atom1.z*((Atom4.z-Atom2.z)*v.x-(Atom4.x-Atom2.x)*v.z);
        hessian-=2.0*Atom1.y*((Atom4.x-Atom2.x)*v.y-(Atom4.y-Atom2.y)*v.x);
        return hessian;
}

Real calcRestraintHessian(AtomStruct &mobileAtom, AtomStruct &restraintAtom, AtomStruct &bondAtom1, AtomStruct &bondAtom2, VectorStruct &bond1, VectorStruct &bond2)
{
        Real sum=0;
        VectorStruct v24, v13;

        v24=atomVector(mobileAtom, bondAtom2);
        v13=atomVector(restraintAtom, bondAtom1);

        sum=(bond2.z*v24.y-bond2.y*v24.z)*(bond1.z*v13.y-bond1.y*v13.z);
        sum+=(bond2.x*v24.z-bond2.z*v24.x)*(bond1.x*v13.z-bond1.z*v13.x);
        sum+=(bond2.y*v24.x-bond2.x*v24.y)*(bond1.y*v13.x-bond1.x*v13.y);
        sum*=2.0;
        return sum;
}

Real calcLocalMoveHessian(ProteinStruct &Protein, vector<AtomStruct> &restraintAtoms, vector<int> &restraintAtomsIndex, vector<int> &variableDihedrals, int &atom1, VectorStruct &bond1, int &index1, int &index2)
{
        //int nrestraint=restraintAtoms.size();
        int atom3, atom4;
        Real sum=0;
        VectorStruct bond2;
        //allignBondWithXAxis(Protein, restraintAtoms, variableDihedrals[index1]);
        getBondAtoms(Protein.Atoms, variableDihedrals[index2], atom3, atom4);
        bond2=atomVector(Protein.Atoms[atom4], Protein.Atoms[atom3]);
        vectorNormalize(bond2);
        for (int i=0;i<2;i++)
        {
                //sum1+=calcRestraintHessian(restraintAtoms[i], Protein.Atoms[atom3], Protein.Atoms[atom4], Protein.Atoms[restraintAtomsIndex[i]]);
                sum+=calcRestraintHessian(Protein.Atoms[restraintAtomsIndex[i]], restraintAtoms[i], Protein.Atoms[atom1], Protein.Atoms[atom3], bond1, bond2);
        }
        //cout <<"hessian1= "<<sum1<<" hessian2= "<<sum2<<endl;
        return sum;
}
/*
Real calcLocalMoveHessian(ProteinStruct &Protein, vector<AtomStruct> &restraintAtoms, vector<int> &restraintAtomsIndex, vector<int> variableDihedrals, int index1, int index2)
{
        //int nrestraint=restraintAtoms.size();
        int atom1, atom2;
        Real sum=0;
        //allignBondWithXAxis(Protein, restraintAtoms, variableDihedrals[index1]);
        getBondAtoms(Protein.Atoms, variableDihedrals[index2], atom1, atom2);
        for (int i=0;i<2;i++)
        {
                sum+=calcRestraintHessian(restraintAtoms[i], Protein.Atoms[atom1], Protein.Atoms[atom2], Protein.Atoms[restraintAtomsIndex[i]]);
        }
        return sum;
}
*/
Real calcRestraintSecondDerivative(AtomStruct &Atom1, AtomStruct &Atom2)
{
        Real y1, z1;
        Real y2, z2;

        y1=Atom1.y;
        z1=Atom1.z;

        y2=Atom2.y;
        z2=Atom2.z;

        return 2.0*(y1*y2+z1*z2);
}

Real calcRestraintSecondDerivative(AtomStruct &mobileAtom, AtomStruct &restraintAtom, AtomStruct &bondAtom, VectorStruct &bond)
{
        Real sum=0;
        Real vyz, vxz, vxy;
        Real bondxy, bondxz, bondyz;
        VectorStruct v12, v23;

        v12=atomVector(mobileAtom, restraintAtom);
        v23=atomVector(restraintAtom, bondAtom);

        vyz=bond.z*v23.y-bond.y*v23.z;
        vxz=bond.x*v23.z-bond.z*v23.x;
        vxy=bond.y*v23.x-bond.x*v23.y;

        bondxy=bond.x*bond.y;
        bondxz=bond.x*bond.z;
        bondyz=bond.y*bond.z;
        
        sum=vyz*vyz;
        sum+=v12.x*(v23.x*(bond.x*bond.x-1.0)+v23.y*bondyz+v23.z*bondxz);

        sum+=vxz*vxz;
        sum+=v12.y*(v23.y*(bond.y*bond.y-1.0)+v23.z*bondyz+v23.x*bondxy);

        sum+=vxy*vxy;
        sum+=v12.z*(v23.z*(bond.z*bond.z-1.0)+v23.x*bondxz+v23.y*bondyz);
        sum*=2.0;
        return sum;        
}

Real calcLocalMoveSecondDerivative(ProteinStruct &Protein, vector<AtomStruct> &restraintAtoms, vector<int> &restraintAtomsIndex, vector<int> &variableDihedrals, int atom1, VectorStruct &bond, int &index)
{
        //int nrestraint=restraintAtoms.size();
        Real sum1=0, sum2=0;
        //allignBondWithXAxis(Protein, restraintAtoms, variableDihedrals[index]);
        for (int i=0;i<2;i++)
        {
                sum1+=calcRestraintSecondDerivative(Protein.Atoms[restraintAtomsIndex[i]], restraintAtoms[i]);
                sum2+=calcRestraintSecondDerivative(Protein.Atoms[restraintAtomsIndex[i]], restraintAtoms[i], Protein.Atoms[atom1], bond);
        }
        cout <<"SecondDerivative1= "<<sum1<<" SecondDerivative2= "<<sum2<<endl;
        return sum2;
}

void calcLocalMoveHessian(ProteinStruct &Protein, vector<AtomStruct> &restraintAtoms, vector<int> &restraintAtomsIndex, vector<int> &variableDihedrals, Real &derivative, Matrix &hessian, int &index)
{
        int ndihedral=variableDihedrals.size();
        int atom1, atom2;
        VectorStruct bond;
        //allignBondWithXAxis(Protein, restraintAtoms, variableDihedrals[index]);
        getBondAtoms(Protein.Atoms, variableDihedrals[index], atom1, atom2);
        bond=atomVector(Protein.Atoms[atom2], Protein.Atoms[atom1]);
        vectorNormalize(bond);
        derivative=calcRestraintDerivative(Protein, restraintAtoms, restraintAtomsIndex, variableDihedrals[index], atom1, bond, index);
        //hessian[index][index]=calcLocalMoveSecondDerivative(Protein, restraintAtoms, restraintAtomsIndex, variableDihedrals, atom1, bond, index);
        hessian[index][index]=calcLocalMoveHessian(Protein, restraintAtoms, restraintAtomsIndex, variableDihedrals, atom1, bond, index, index);
        for (int i=index+1;i<ndihedral;i++)
        {
                hessian[index][i]=calcLocalMoveHessian(Protein, restraintAtoms, restraintAtomsIndex, variableDihedrals, atom1, bond, index, i);
                hessian[i][index]=hessian[index][i];
        }
}

void calcLocalMoveHessian(ProteinStruct &Protein, vector<AtomStruct> &restraintAtoms, vector<int> &restraintAtomsIndex, vector<int> &variableDihedrals, vector<Real> &gradient, Matrix &hessian)
{
        int ndihedral=variableDihedrals.size();
        for (int i=0;i<ndihedral;i++)
        {
                calcLocalMoveHessian(Protein, restraintAtoms, restraintAtomsIndex, variableDihedrals, gradient[i], hessian, i);
        }
}

vector<AtomStruct> truncateProtein(vector<AtomStruct> &Atoms, int residue1, int residue2)
{
        int natom=Atoms.size();
        vector<AtomStruct> truncatedAtoms;

        for (int i=0;i<natom;i++)
        {
                if (Atoms[i].ResidueNum>=residue1 && Atoms[i].ResidueNum<=residue2 && (Atoms[i].AtomName=="CA" || Atoms[i].AtomName=="C" || Atoms[i].AtomName=="N"))
                {
                        SafePushBack(truncatedAtoms, Atoms[i], "truncatedAtoms");
                }
        }
        renumberResidues(truncatedAtoms);
	setDefaultConnectivity(truncatedAtoms);
        return truncatedAtoms;
}

vector<AtomStruct> truncateProtein(vector<AtomStruct> &Atoms, int &residue)
{
        int natom=Atoms.size();
        vector<AtomStruct> truncatedAtoms;

        for (int i=0;i<natom;i++)
        {
                if (Atoms[i].ResidueNum<=residue && (Atoms[i].AtomName=="CA" || Atoms[i].AtomName=="C" || Atoms[i].AtomName=="N"))
                {
                        SafePushBack(truncatedAtoms, Atoms[i], "truncatedAtoms");
                }
        }
	setDefaultConnectivity(truncatedAtoms);
        return truncatedAtoms;
}

void rotateToAllign(vector<AtomStruct> &Atoms, AtomStruct &restraintAtom, int &atomIndex, int &atom1, int &atom2)
{
        Real angle;

        angle=calcDihedralAngle(Atoms[atomIndex], Atoms[atom1], Atoms[atom2], restraintAtom);
	if (!isnan(angle)) rotateAtomsAroundVector(Atoms, atom1, atom2, angle);
}

void rotateToAllign(vector<AtomStruct> &Atoms, AtomStruct restraintAtom2, AtomStruct restraintAtom3, int &atomIndex, int &atom1, int &atom2)
{
        Real angle;
	Real dx, dy, dz;
	
	dx=restraintAtom2.x-Atoms[atom2].x;
	dy=restraintAtom2.y-Atoms[atom2].y;
	dz=restraintAtom2.z-Atoms[atom2].z;

	moveAtom(restraintAtom3, -dx, -dy, -dz);

        angle=calcDihedralAngle(Atoms[atomIndex], Atoms[atom1], Atoms[atom2], restraintAtom3);
	if (!isnan(angle)) rotateAtomsAroundVector(Atoms, atom1, atom2, angle);
}

Real optimizeAngle(ProteinStruct &Protein, vector<AtomStruct> &restraintAtoms, vector<int> &restraintAtomsIndex, vector<int> &variableDihedrals, int &index)
{
        Real numerator=0, denominator=0, angle;
        allignBondWithXAxis(Protein, restraintAtoms, variableDihedrals[index]);
        for (int i=0;i<2;i++)
        {
                numerator+=(Protein.Atoms[restraintAtomsIndex[i]].z*restraintAtoms[i].y-restraintAtoms[i].z*Protein.Atoms[restraintAtomsIndex[i]].y);
                denominator+=(Protein.Atoms[restraintAtomsIndex[i]].y*restraintAtoms[i].y+restraintAtoms[i].z*Protein.Atoms[restraintAtomsIndex[i]].z);
        }
        angle=atan(numerator/denominator);
        return angle;
}

Real optimizeAnglesIndividually(ProteinStruct &Protein, vector<AtomStruct> &restraintAtoms, vector<Real> &dihedrals, vector<int> &restraintAtomsIndex, vector<int> &variableDihedrals)
{
        int atom1, atom2;
        int ndihedral=variableDihedrals.size();
        Real angle, score=UNK_REAL, score_old;
        score_old=calcRestraintEnergy(Protein, restraintAtoms, restraintAtomsIndex);
        for (int i=0;i<ndihedral;i++)
        {
                angle=optimizeAngle(Protein, restraintAtoms, restraintAtomsIndex, variableDihedrals, i);
                getBondAtoms(Protein.Atoms, variableDihedrals[i], atom1, atom2);
                if (!isnan(angle))
                {
                rotateAtomsAroundVector(Protein.Atoms, atom1, atom2, -angle);
                dihedrals[i]-=angle;
		//cout <<"AngleStep= "<<angle<<endl;
                score=calcRestraintEnergy(Protein, restraintAtoms, restraintAtomsIndex);
                if (score>score_old) 
                {
                        rotateAtomsAroundVector(Protein.Atoms, atom1, atom2, pi);
                        dihedrals[i]+=pi;
                        score=calcRestraintEnergy(Protein, restraintAtoms, restraintAtomsIndex);
                }
                }
                //cout <<"score= "<<score<<" score_old= "<<score_old<<endl;
                score_old=score;
        }
        //cout <<endl;
        return score;
}

/*
Real localMoveHessian(ProteinStruct &Protein, int residue, int angleType, Real dAngle)
{
        int iterations=10;
        int lastRestraintIndex, lastDihedral;
        Real dScoreCriteria=10e-9, criteria=10e-8, score=0, score_old=0;
        vector<int> restraintAtomsIndex, variableDihedrals;
        vector<Real> gradient, dihedrals, step;
        AtomStruct lastRestraintAtom;
        Matrix hessian;
        vector<AtomStruct> restraintAtoms;
        MinimizeRestraintStruct args;
        Real (*func_to_minimize)(vector<Real>, MinimizeRestraintStruct)=calcRestraintEnergy;
        ProteinStruct truncatedProtein;
        //cout <<"About to getRestraints"<<endl;
        getRestraintAtoms(Protein.Atoms, restraintAtoms, restraintAtomsIndex, residue, angleType);
        //cout <<"About to getVariableDihedrals"<<endl;
        variableDihedrals=getVariableDihedrals(residue);
        lastRestraintIndex=getAtomIndex(Protein.Atoms, Protein.Atoms[restraintAtomsIndex[1]].ResidueNum+1, "N");
        lastDihedral=variableDihedrals[int(variableDihedrals.size())-1]+1;
        lastRestraintAtom=Protein.Atoms[lastRestraintIndex];
        pivot(Protein.Atoms, residue, angleType, dAngle);
        args.Protein=Protein;
        args.restraintAtoms=restraintAtoms;
        args.restraintAtomsIndex=restraintAtomsIndex;
        args.variableDihedrals=variableDihedrals;
        calcDihedrals(args.Protein.Atoms, args.Protein.dihedral);
        setVariableDihedrals(dihedrals, variableDihedrals, args.Protein.dihedral);
        
        int ndihedral=variableDihedrals.size();
        //cout <<"In calcLocalMoveHessian"<<endl;
        Safe2DAlloc(hessian, ndihedral, ndihedral, "hessian");
        optimizeAnglesIndividually(Protein, restraintAtoms, restraintAtomsIndex, variableDihedrals);
        //score_old=calcRestraintEnergy(dihedrals, Protein, restraintAtoms, restraintAtomsIndex, variableDihedrals);
        for (int i=0;i<iterations;i++)
        {
                //args.Protein=truncatedProtein;
                args.Protein=Protein;
                calcDihedrals(args.Protein.Atoms, args.Protein.dihedral);
                setVariableDihedrals(dihedrals, variableDihedrals, args.Protein.dihedral);
                gradient=calcLocalMoveGradient(Protein, restraintAtoms, restraintAtomsIndex, variableDihedrals); 
                //PrintPdb("/home/jouko/project/Density/pdb/debug.pdb", Protein.Atoms);
                //PrintPdb("/home/jouko/project/Density/pdb/restraint.pdb", restraintAtoms);
                calcLocalMoveHessian(Protein, restraintAtoms, restraintAtomsIndex, variableDihedrals, hessian);
                //printMatrix(hessian);
                //if (isPositiveDefinite(hessian))
                //{
                        step=solveHessian(gradient, hessian);
                        for (int j=0;j<ndihedral;j++)
                        {
                                dihedrals[j]+=step[j];
                        }
                        //score=MinimizeAlongVector(dihedrals, step, args, func_to_minimize);
                //}
                //else
                //{
                //       scaleVector(gradient, 0.1);
                //        score=MinimizeAlongVector(dihedrals, gradient, args, func_to_minimize);
                //}
                for (int j=0;j<int(dihedrals.size());j++)
                {
                        setDihedralAngle(Protein.Atoms, variableDihedrals[j], dihedrals[j]);
                }
                score=calcRestraintEnergy(Protein, restraintAtoms, restraintAtomsIndex);
                score_old=score;
                if (score<criteria) break;
        }
        //PrintPdb("/home/jouko/project/Density/pdb/debug.pdb", Protein.Atoms);
        //PrintPdb("/home/jouko/project/Density/pdb/restraint.pdb", restraintAtoms);
        rotateToAllign(Protein.Atoms, lastRestraintAtom, lastRestraintIndex, restraintAtomsIndex[0], restraintAtomsIndex[1]);
        calcDihedrals(Protein.Atoms, Protein.dihedral);
        //PrintPdb("/home/jouko/project/Density/pdb/debug.pdb", Protein.Atoms);
        //PrintPdb("/home/jouko/project/Density/pdb/restraint.pdb", restraintAtoms);
        cout <<"loclaMoveFit= "<<score<<endl;
}
*/

Real localMoveHessian(ProteinStruct &Protein, int residue, int angleType, Real dAngle)
{
        int iterations=100, lastRestraintIndex;
        Real criteria=10e-10, score=0, score_old=0;
        vector<int> restraintAtomsIndex, variableDihedrals;
        vector<Real> gradient, dihedrals, step;
        AtomStruct lastRestraintAtom;
        Matrix hessian;
        vector<AtomStruct> restraintAtoms;
        getRestraintAtoms(Protein.Atoms, restraintAtoms, restraintAtomsIndex, residue, angleType);
        variableDihedrals=getVariableDihedrals(residue);
        lastRestraintIndex=getAtomIndex(Protein.Atoms, Protein.Atoms[restraintAtomsIndex[1]].ResidueNum+1, "N");
        lastRestraintAtom=Protein.Atoms[lastRestraintIndex];
        //pivot(Protein.Atoms, residue, angleType, dAngle);
        doubleCrank(Protein.Atoms, residue, angleType, dAngle);
        calcDihedrals(Protein.Atoms, Protein.dihedral);
        setVariableDihedrals(dihedrals, variableDihedrals, Protein.dihedral);
        int ndihedral=variableDihedrals.size();
        //cout <<"In calcLocalMoveHessian"<<endl;
        Safe2DAlloc(hessian, ndihedral, ndihedral, "hessian");
        SafeAlloc(gradient, ndihedral, "gradient");
        score_old=optimizeAnglesIndividually(Protein, restraintAtoms, dihedrals, restraintAtomsIndex, variableDihedrals);
	//cout <<"score_old= "<<score_old<<endl;
        for (int i=0;i<iterations;i++)
        {
                //gradient=calcLocalMoveGradient(Protein, restraintAtoms, restraintAtomsIndex, variableDihedrals); 
                calcLocalMoveHessian(Protein, restraintAtoms, restraintAtomsIndex, variableDihedrals, gradient, hessian);
                step=solveHessian(gradient, hessian);
                for (int j=0;j<ndihedral;j++)
                {
                        dihedrals[j]+=step[j];
                        if (!isnan(step[j]) && !isinf(step[j]))
                        {
                                rotateDihedralAngleBy(Protein.Atoms, variableDihedrals[j], step[j]);
                        }
                }
                score=calcRestraintEnergy(Protein, restraintAtoms, restraintAtomsIndex);
		//if (score>score_old) break;
                if (score>score_old)
                {
                        for (int j=0;j<ndihedral;j++)
                        {
                                dihedrals[j]-=step[j];
                                setDihedralAngle(Protein.Atoms, variableDihedrals[j], dihedrals[j]);
                        }
                        score=optimizeAnglesIndividually(Protein, restraintAtoms, dihedrals, restraintAtomsIndex, variableDihedrals);
			//cout <<"score= "<<score<<endl;
                }
                score_old=score;
                if (score<criteria) break;
        }
        rotateToAllign(Protein.Atoms, restraintAtoms[1], restraintAtoms[2], lastRestraintIndex, restraintAtomsIndex[0], restraintAtomsIndex[1]);
	calcDihedrals(Protein.Atoms, Protein.dihedral);
        return score;
}

Real localMoveHessian(ProteinStruct &Protein, vector<int> &restraintAtomsIndex, vector<int> variableDihedrals, int lastRestraintIndex, int lastDihedral, vector<AtomStruct> &restraintAtoms, AtomStruct &lastRestraintAtom)
{
        int iterations=100;
        Real criteria=10e-10, score=0, score_old=0;
        vector<Real> gradient, dihedrals, step;
        Matrix hessian;
        calcDihedrals(Protein.Atoms, Protein.dihedral);
        setVariableDihedrals(dihedrals, variableDihedrals, Protein.dihedral);
        int ndihedral=variableDihedrals.size();
        Safe2DAlloc(hessian, ndihedral, ndihedral, "hessian");
        SafeAlloc(gradient, ndihedral, "gradient");
        score_old=optimizeAnglesIndividually(Protein, restraintAtoms, dihedrals, restraintAtomsIndex, variableDihedrals);
	//cout <<"score_old= "<<score_old<<endl;
        for (int i=0;i<iterations;i++)
        {
                //gradient=calcLocalMoveGradient(Protein, restraintAtoms, restraintAtomsIndex, variableDihedrals); 
                calcLocalMoveHessian(Protein, restraintAtoms, restraintAtomsIndex, variableDihedrals, gradient, hessian);
                step=solveHessian(gradient, hessian);
                for (int j=0;j<ndihedral;j++)
                {
                        dihedrals[j]+=step[j];
                        if (!isnan(step[j]) && !isinf(step[j]))
                        {
                                rotateDihedralAngleBy(Protein.Atoms, variableDihedrals[j], step[j]);
                        }
                }
                score=calcRestraintEnergy(Protein, restraintAtoms, restraintAtomsIndex);
		//if (score>score_old) break;
                if (score>score_old)
                {
                        for (int j=0;j<ndihedral;j++)
                        {
                                dihedrals[j]-=step[j];
                                setDihedralAngle(Protein.Atoms, variableDihedrals[j], dihedrals[j]);
                        }
                        score=optimizeAnglesIndividually(Protein, restraintAtoms, dihedrals, restraintAtomsIndex, variableDihedrals);
			//cout <<"score= "<<score<<endl;
                }
                score_old=score;
                if (score<criteria) break;
        }
        rotateToAllign(Protein.Atoms, restraintAtoms[1], restraintAtoms[2], lastRestraintIndex, restraintAtomsIndex[0], restraintAtomsIndex[1]);
	calcDihedrals(Protein.Atoms, Protein.dihedral);
        return score;
}

Real localMoveHessianVeryFast(ProteinStruct &Protein, int residue, int angleType, Real dAngle)
{
        int ndihedral, deletedDihedrals, truncatedResidue;
        int lastRestraintIndex, lastDihedral;
	int firstResidue;
        vector<int> restraintAtomsIndex, variableDihedrals;
        Real localMoveFit;
	AtomStruct lastRestraintAtom;
        vector<AtomStruct> restraintAtoms;
        ProteinStruct truncatedProtein;
        
	getFirstResidue(Protein.Atoms, firstResidue);
        if (residue>1) truncatedResidue=2;
        else truncatedResidue=1;
        truncatedProtein.Atoms=truncateProtein(Protein.Atoms, residue-1, residue+5);
	getRestraintAtoms(truncatedProtein.Atoms, restraintAtoms, restraintAtomsIndex, truncatedResidue, angleType);
        variableDihedrals=getVariableDihedrals(truncatedResidue);
        lastRestraintIndex=getAtomIndex(truncatedProtein.Atoms, truncatedProtein.Atoms[restraintAtomsIndex[1]].ResidueNum+1, "N");
        lastDihedral=variableDihedrals[int(variableDihedrals.size())-1]+1;
        lastRestraintAtom=truncatedProtein.Atoms[lastRestraintIndex];
        doubleCrank(truncatedProtein.Atoms, truncatedResidue, angleType, dAngle);
	localMoveFit=localMoveHessian(truncatedProtein, restraintAtomsIndex, variableDihedrals, lastRestraintIndex, lastDihedral, restraintAtoms, lastRestraintAtom);
        ndihedral=truncatedProtein.dihedral.size();
        if (residue>firstResidue+1) deletedDihedrals=(residue-2)*3;
        else deletedDihedrals=0;
        for (int i=0;i<ndihedral;i++)
        {
                setDihedralAngle(Protein.Atoms, i+deletedDihedrals, truncatedProtein.dihedral[i]);
        }
        return localMoveFit;

}


