/*
    JV Library
    Copyright (C) 2014  Jouko Juhani Virtanen

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
# include <iostream>
# include <algorithm>

using namespace std;

# include "IOUtils.h"
# include "Structures.h"
# include "TypeDef.h"
# include "StringUtils.h"



void printSymmetryOperations(vector<SymmetryOperationStruct> &symmetryOperations)
{
	int noperations=symmetryOperations.size();

	for (int i=0;i<noperations;i++)
	{
		for (int j=0;j<3;j++)
		{
			cout <<"REMARK 290   SMTRY"<<j+1<<"   "<<i+1<<"  ";
			cout <<setprecision(6)
			<<symmetryOperations[i].rotationMat[j][X]<<"  "
			<<symmetryOperations[i].rotationMat[j][Y]<<"  "
			<<symmetryOperations[i].rotationMat[j][Z]<<"  ";
			cout <<setprecision(5)<<setw(15)
			<<setiosflags(ios::right)
			<<symmetryOperations[i].translationVector[j]<<endl;
		}
	}
}

void printSymmetryOperations(string PdbFile, vector<SymmetryOperationStruct> &symmetryOperations)
{
	ofstream file;
	int noperations=symmetryOperations.size();

	OpenFileForWriting(PdbFile, file);

	for (int i=0;i<noperations;i++)
	{
		for (int j=0;j<3;j++)
		{
			file <<"REMARK 290   SMTRY"<<j+1<<"   "<<i+1;
			file <<setw(10)<<setprecision(6)
			<<setiosflags(ios::fixed)
			<<setiosflags(ios::right)
			<<symmetryOperations[i].rotationMat[j][X];
			file <<setw(10)<<setprecision(6)
			<<setiosflags(ios::fixed)
			<<setiosflags(ios::right)
			<<symmetryOperations[i].rotationMat[j][Y];
			file <<setw(10)<<setprecision(6)
			<<setiosflags(ios::fixed)
			<<setiosflags(ios::right)
			<<symmetryOperations[i].rotationMat[j][Z];
			file <<setprecision(5)<<setw(15)
			<<setiosflags(ios::right)
			<<symmetryOperations[i].translationVector[j]<<endl;
		}
	}
	file.close();
}

void printCryst(string PdbFile, ProteinStruct &Protein)
{
	ofstream file;
	string spaceGroup;
	int length, numSpaces;

	if (Protein.spaceGroup=="A1") spaceGroup="A 1";
	else if (Protein.spaceGroup=="A2") spaceGroup="A 2";
	else if (Protein.spaceGroup=="B112") spaceGroup="B 1 1 2";
	else if (Protein.spaceGroup=="B2") spaceGroup="B 2";
	else if (Protein.spaceGroup=="B2212") spaceGroup="B 2 21 2";
	else if (Protein.spaceGroup=="C121") spaceGroup="C 1 2 1";
	else if (Protein.spaceGroup=="C222") spaceGroup="C 2 2 2";
	else if (Protein.spaceGroup=="C2221") spaceGroup="C 2 2 21";
	else if (Protein.spaceGroup=="C21") spaceGroup="C 21";
	else if (Protein.spaceGroup=="C4212") spaceGroup="C 4 21 2";
	else if (Protein.spaceGroup=="F222") spaceGroup="F 2 2 2";
	else if (Protein.spaceGroup=="F23") spaceGroup="F 2 3";
	else if (Protein.spaceGroup=="F422") spaceGroup="F 4 2 2";
	else if (Protein.spaceGroup=="F432") spaceGroup="F 4 3 2";
	else if (Protein.spaceGroup=="F4132") spaceGroup="F 41 3 2";
	else if (Protein.spaceGroup=="H3") spaceGroup="H 3";
	else if (Protein.spaceGroup=="H32") spaceGroup="H 3 2";
	else if (Protein.spaceGroup=="I42d") spaceGroup="I -4 2 d";
	else if (Protein.spaceGroup=="I4c2") spaceGroup="I -4 c 2";
	else if (Protein.spaceGroup=="I121") spaceGroup="I 1 2 1";
	else if (Protein.spaceGroup=="I1211") spaceGroup="I 1 21 1";
	else if (Protein.spaceGroup=="I222") spaceGroup="I 2 2 2";
	else if (Protein.spaceGroup=="I23") spaceGroup="I 2 3";
	else if (Protein.spaceGroup=="I21") spaceGroup="I 2 1";
	else if (Protein.spaceGroup=="I212121") spaceGroup="I 21 21 21";
	else if (Protein.spaceGroup=="I213") spaceGroup="I 21 3";
	else if (Protein.spaceGroup=="I4") spaceGroup="I 4";
	else if (Protein.spaceGroup=="I422") spaceGroup="I 4 2 2";
	else if (Protein.spaceGroup=="I432") spaceGroup="I 4 3 2";
	else if (Protein.spaceGroup=="I41a") spaceGroup="I 41/a";
	else if (Protein.spaceGroup=="P_1") spaceGroup="P -1";
	else if (Protein.spaceGroup=="P 1") spaceGroup="P 1";
	else if (Protein.spaceGroup=="P112") spaceGroup="P 1 1 2";
	else if (Protein.spaceGroup=="P1121") spaceGroup="P 1 1 21";
	else if (Protein.spaceGroup=="P121") spaceGroup="P 1 2 1";
	else if (Protein.spaceGroup=="P1211") spaceGroup="P 1 21 1";
	else if (Protein.spaceGroup=="P121c1") spaceGroup="P 1 21/c 1";
	else if (Protein.spaceGroup=="P121n1") spaceGroup="P 1 21/n 1";
	else if (Protein.spaceGroup=="P222") spaceGroup="P 2 2 2";
	else if (Protein.spaceGroup=="P2221") spaceGroup="P 2 2 21";
	else if (Protein.spaceGroup=="P22121") spaceGroup="P 2 21 21";
	else if (Protein.spaceGroup=="P21212") spaceGroup="P 21 21 2";
	else if (Protein.spaceGroup=="P21212A") spaceGroup="P 21 21 2 A";
	else if (Protein.spaceGroup=="P212121") spaceGroup="P 21 21 21";
	else if (Protein.spaceGroup=="P213") spaceGroup="P 2 1 3";
	else if (Protein.spaceGroup=="P3") spaceGroup="P 3";
	else if (Protein.spaceGroup=="P312") spaceGroup="P 3 1 2";
	else if (Protein.spaceGroup=="P321") spaceGroup="P 3 2 1";
	else if (Protein.spaceGroup=="P31") spaceGroup="P 3 1";
	else if (Protein.spaceGroup=="P3112") spaceGroup="P 1 1 2";
	else if (Protein.spaceGroup=="P32") spaceGroup="P 32";
	else if (Protein.spaceGroup=="P3212") spaceGroup="P 32 1 2";
	else if (Protein.spaceGroup=="P3221") spaceGroup="P 32 2 1";
	else if (Protein.spaceGroup=="P4") spaceGroup="P 4";
	else if (Protein.spaceGroup=="P422") spaceGroup="P 4 2 2";
	else if (Protein.spaceGroup=="P4212") spaceGroup="P 4 21 2";
	else if (Protein.spaceGroup=="P4132") spaceGroup="P 41 3 2";
	else if (Protein.spaceGroup=="P42") spaceGroup="P 42";
	else if (Protein.spaceGroup=="P4222") spaceGroup="P 42 2 2";
	else if (Protein.spaceGroup=="P4332") spaceGroup="P 43 3 2";
	else if (Protein.spaceGroup=="P6") spaceGroup="P 6";
	else if (Protein.spaceGroup=="P61") spaceGroup="P 61";
	else if (Protein.spaceGroup=="P6122") spaceGroup="P 61 2 2";
	else if (Protein.spaceGroup=="P63") spaceGroup="P 63";
	else if (Protein.spaceGroup=="P6322") spaceGroup="P 63 2 2";
	else if (Protein.spaceGroup=="P64") spaceGroup="P 64";
	else if (Protein.spaceGroup=="P6422") spaceGroup="P 64 2 2";
	else if (Protein.spaceGroup=="P6522") spaceGroup="P 65 2 2";
	else if (Protein.spaceGroup=="R3") spaceGroup="R 3";
	else if (Protein.spaceGroup=="R32") spaceGroup="R 32";
	else spaceGroup=Protein.spaceGroup;
	if (Protein.XBoxLength!=0 && Protein.YBoxLength!=0)
	{
		OpenFileForWriting(PdbFile, file);

		file <<"CRYST1";
		file <<setw(9)<<setiosflags(ios::right)<<setprecision(3)
		<<setiosflags(ios::fixed)
		<<Protein.XBoxLength;
		file <<setw(9)<<setiosflags(ios::right)<<setprecision(3)
		<<Protein.YBoxLength;
		file <<setw(9)<<setiosflags(ios::right)<<setprecision(3)
		<<Protein.ZBoxLength;
		file <<setiosflags(ios::right)<<setprecision(2)<<setw(7)
		<<Protein.alpha;
		file <<setiosflags(ios::right)<<setprecision(2)<<setw(7)
		<<Protein.beta;
		file <<setiosflags(ios::right)<<setprecision(2)<<setw(7)
		<<Protein.gamma;
		//file <<setw(3)<<Protein.bravisLatticeCentering<<Protein.symmetry1
		//<<Protein.symmetry2<<Protein.symmetry3<<setw(5)
		file <<" ";
		file << resetiosflags(ios::left);
		file <<resetiosflags(ios::left)<<spaceGroup;
		length=spaceGroup.length();
		numSpaces=14-length;
		for (int i=0;i<numSpaces;i++) file <<" ";
		file <<Protein.symmetryOperations.size()<<endl;

		file.close();
	}
}

void printScale(string PdbFile, Matrix &cartToFrac)
{
	ofstream file;
	int Size=cartToFrac.size();

	if (Size==3)
	{
		OpenFileForWriting(PdbFile, file);
		for (int i=0;i<3;i++)
		{
			file <<"SCALE"<<i+1<<"    ";
			for (int j=0;j<3;j++)
			{
				file <<setiosflags(ios::right)<<setprecision(6)
				<<setw(10)<<setiosflags(ios::fixed)<<cartToFrac[i][j];
			}
			file <<"        0.00000"<<endl;
		}
		file.close();
	}
}

void printConnectivity(string pdbFile, vector<AtomStruct> &Atoms)
{
	ofstream pdb;
	int natom=Atoms.size();
	int nconnect;

	OpenFileForWriting(pdbFile, pdb);
	for (int i=0;i<natom;i++)
	{
		nconnect=Atoms[i].connectivity.size();
		if (nconnect>0)
		{
			pdb <<"CONECT";
			pdb <<setiosflags(ios::right)<<setw(5)<<i+1;
			for (int j=0;j<nconnect;j++)
			{
				pdb <<setiosflags(ios::right)<<setw(5)<<Atoms[i].connectivity[j];
			}
			pdb <<endl;
		}
	}
}

void printPdbLine(ofstream &pdb, bool HetAtom, int AtomNumber, string AtomName, string ResidueName, int ResidueNum, string ChainName, Real x, Real y, Real z, Real Occupancy, Real BFactor, string segID)
{
	int chars, AtomChars;

	ResidueName=ResidueName+".";
	AtomName=AtomName+".";

	chars=ResidueName.find('.');
	AtomChars=AtomName.find('.');
        if (ResidueNum>9999) ResidueNum=9999;
	ResidueName=ResidueName.substr(0,chars);
	AtomName=AtomName.substr(0,AtomChars)+" ";
	
	pdb << setfill(' ')
	        << setw(6) << setiosflags(ios::left) << (HetAtom ? "HETATM" : "ATOM") << resetiosflags(ios::left)
		<< setw(5) << setiosflags(ios::right) << (AtomNumber>100000 ? hex : dec ) << AtomNumber << resetiosflags(ios::right)
		<< (AtomChars>3 ? " " : "  ")
		<< (AtomChars>3 ? setw(3) : setw(4))
		<< (AtomChars>3 ? setiosflags(ios::right) : setiosflags(ios::left))
		<< AtomName << resetiosflags(ios::left)
		<< setiosflags(ios::right);

	if (chars==2)
	{
		pdb <<setfill(' ')
			<< setiosflags(ios::right) << " " 
			<< ResidueName << resetiosflags(ios::left);
	}
	if (chars==3 || chars==4)
	{
		pdb <<setfill(' ')
			<< setiosflags(ios::right) << ResidueName 
			<< resetiosflags(ios::left);
	}
	if (chars>3) pdb << setw(1) <<ChainName<< resetiosflags(ios::right);
	else pdb <<setw(1) <<" "<<ChainName<< resetiosflags(ios::right);

	pdb << setw(4) << setiosflags(ios::right) << ResidueNum<<resetiosflags(ios::right);
	pdb	<< setw(4) << " ";
	pdb	<< setiosflags(ios::fixed);
	pdb	<< setw(8) << setprecision(3);
	pdb	<< x;
	pdb	<< setw(8) << setprecision(3);
	pdb	<< y;
	pdb	<< setw(8) << setprecision(3);
	pdb	<< z;
	pdb	<< setw(6) << setprecision(2);
	pdb	<< Occupancy;
	pdb	<< setw(9) << setprecision(5);
	pdb	<< BFactor;
	pdb	<< setw(3) << "   ";
	pdb	<< segID;
	pdb	<< resetiosflags(ios::fixed);
	pdb	<< resetiosflags(ios::right);
	pdb	<< setw(0) <<endl;
	pdb 	<<dec;              
}

void printAnisou(ofstream &pdb, AtomStruct &Atom)
{
	int chars, AtomChars;

	Atom.ResidueName=Atom.ResidueName+".";
	Atom.AtomName=Atom.AtomName+".";

	chars=Atom.ResidueName.find('.');	
	AtomChars=Atom.AtomName.find('.');
        if (Atom.ResidueNum>9999) Atom.ResidueNum=9999;
	Atom.ResidueName=Atom.ResidueName.substr(0,chars);
	Atom.AtomName=Atom.AtomName.substr(0,AtomChars)+" ";
	
	pdb << setfill(' ')
	        << setw(6) << setiosflags(ios::left) << "ANISOU" << resetiosflags(ios::left)
		<< setw(5) << setiosflags(ios::right) << (Atom.AtomNumber>100000 ? hex : dec ) << Atom.AtomNumber << resetiosflags(ios::right)
		<< (AtomChars>3 ? " " : "  ")
		<< (AtomChars>3 ? setw(3) : setw(4))
		<< (AtomChars>3 ? setiosflags(ios::right) : setiosflags(ios::left))
		<< Atom.AtomName << resetiosflags(ios::left)
		<< setiosflags(ios::right);

	if (chars==2)
	{
		pdb <<setfill(' ')
			<< setiosflags(ios::right) << " " 
			<< Atom.ResidueName << resetiosflags(ios::left);
	}
	if (chars==3 || chars==4)
	{
		pdb <<setfill(' ')
			<< setiosflags(ios::right) << Atom.ResidueName 
			<< resetiosflags(ios::left);
	}
	if (chars>3) pdb << setw(1) <<Atom.ChainName<< resetiosflags(ios::right);
	else pdb <<setw(1) <<" "<<Atom.ChainName<< resetiosflags(ios::right);

	pdb << setw(4) << setiosflags(ios::right) << Atom.ResidueNum<<resetiosflags(ios::right);
	pdb <<"  ";
	pdb <<resetiosflags(ios::right);
	for (int i=0;i<6;i++)
	{
		pdb << setw(7) << int(Atom.anisou[i]*10000.0);
	}
	pdb <<  resetiosflags(ios::left);
	pdb <<endl;
}

void printPdbLine(ofstream &pdb, AtomStruct &Atom)
{
	printPdbLine(pdb, Atom.HetAtom, Atom.AtomNumber, Atom.AtomName, Atom.ResidueName, Atom.ResidueNum, Atom.ChainName, Atom.x, Atom.y, Atom.z, Atom.Occupancy, Atom.BFactor, Atom.SegID);
}

void printAtomToFile(string PdbFile, AtomStruct &Atom)
{
	ofstream pdb;
	OpenFileForWriting(PdbFile, pdb);
        printPdbLine(pdb, Atom);
}

void addAtomsToTrj(string PdbOutputFile, vector<AtomStruct> &Atoms)
{
	int natom, nanisou;
	ofstream pdb;
	natom=Atoms.size();
	cout <<"natom= "<<natom<<endl;
	cout <<"writing "<<PdbOutputFile<<endl;

	OpenFileForWriting(PdbOutputFile, pdb);

	pdb <<"MODEL"<<endl;
	for (int n=0;n<natom;n++)
	{
		printPdbLine(pdb, Atoms[n]);
		nanisou=Atoms[n].anisou.size();
		if (nanisou==6) printAnisou(pdb, Atoms[n]);
	}

	pdb <<"ENDMDL"<<endl;
	pdb.close();

}

void printPdb(string PdbFile, vector<AtomStruct> &Atoms)
{
	AddIndexToFile(PdbFile);
        addAtomsToTrj(PdbFile, Atoms);
}

void printPdb(string PdbFile, ProteinStruct &Protein)
{
	AddIndexToFile(PdbFile);
	printSymmetryOperations(PdbFile, Protein.symmetryOperations);
	printCryst(PdbFile, Protein);
	cout <<"Before printScale"<<endl;
	printScale(PdbFile, Protein.cartToFrac);
	cout <<"Before AddAtomsToTrj"<<endl;
	addAtomsToTrj(PdbFile, Protein.Atoms);		
	printConnectivity(PdbFile, Protein.Atoms);
}

void printPdb(string pdbFile, vector<ProteinStruct> &Proteins)
{
	ofstream pdb;
	int natom, nprot=Proteins.size();
	AddIndexToFile(pdbFile);

	OpenFileForWriting(pdbFile, pdb);
	for (int i=0;i<nprot;i++)
	{
		natom=Proteins[i].Atoms.size();
		for (int j=0;j<natom;j++)
		{
			printPdbLine(pdb, Proteins[i].Atoms[j]);
		}
	}
}

void printAtomInfo(AtomStruct Atom)
{
	int chars, AtomChars;

	chars=Atom.ResidueName.length();
	AtomChars=Atom.AtomName.length();

	Atom.ResidueName=Atom.ResidueName.substr(0,chars);
	Atom.AtomName=Atom.AtomName.substr(0,AtomChars)+" ";
	
	cout  << setfill(' ')
		<< setw(6) << setiosflags(ios::left) << (Atom.HetAtom ? "HETATM" : "ATOM") << resetiosflags(ios::left)
		<< setw(5) << setiosflags(ios::right) << (Atom.AtomNumber>=100000 ? hex : dec ) << Atom.AtomNumber << resetiosflags(ios::right)
		<< (AtomChars>3 ? " " : "  ")
		<< (AtomChars>3 ?  setw(3) : setw(4))
		<< (AtomChars>3 ?  setiosflags(ios::right) : setiosflags(ios::left))
		<< Atom.AtomName << resetiosflags(ios::left)
		//<< " "
		<< setiosflags(ios::right);
	//<< setw(4)
	cout <<setfill(' ')
		<< setiosflags(ios::right)<< Atom.ResidueName <<resetiosflags(ios::left);

	if (chars>3) cout << setw(1) <<"A"<< resetiosflags(ios::right);
	else cout << setw(2) <<" A"<< resetiosflags(ios::right);

	cout << setw(4) << setiosflags(ios::right)<<Atom.ResidueNum<<resetiosflags(ios::right)
		<< setw(4) << " "
		<< setiosflags(ios::fixed)
		<< setw(8) << setprecision(3)
		<< Atom.x
		<< setw(8) << setprecision(3)
		<< Atom.y
		<< setw(8) << setprecision(3)
		<< Atom.z
		<< setw(6) << setprecision(2)
		<< Atom.Occupancy
		<< setw(9) << setprecision(5)
		<< Atom.BFactor
		<< setw(14) << " "
		<< resetiosflags(ios::fixed)
		<< resetiosflags(ios::right)
		<< setw(0) <<endl;         
	cout <<dec<<endl;              
}

void printPdbToScreen(vector<AtomStruct> &Atoms)
{
	int natom=Atoms.size();

	for (int i=0;i<natom;i++)
	{
		printAtomInfo(Atoms[i]);
	}
}
