/*
    JV Library
    Copyright (C) 2014  Jouko Juhani Virtanen

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

void setConnectivity(vector<AtomStruct> &Atoms, int index)
{
	int natom=Atoms.size();

	for (int i=0;i<natom;i++)
	{
		if (Atoms[index].AtomName=="CA")
		{
			if (isSameResidue(Atoms[index], Atoms[i]))
			{
				if (Atoms[i].AtomName=="N" || Atoms[i].AtomName=="C" || Atoms[i].AtomName=="CB")
				{
					SafePushBack(Atoms[index].connectivity, i, "connectivity");
				}
			}
		}
		else if (Atoms[index].AtomName=="CB")
		{
			if (isSameResidue(Atoms[index], Atoms[i]))
			{
				if (Atoms[i].AtomName=="CA")
				{
					SafePushBack(Atoms[index].connectivity, i, "connectivity");
				}
			}
		}
		else if (Atoms[index].AtomName=="C")
		{
			if (isSameResidue(Atoms[index], Atoms[i]))
			{
				if (Atoms[i].AtomName=="CA" || Atoms[i].AtomName=="O")
				{
					SafePushBack(Atoms[index].connectivity, i, "connectivity");
				}
			}
			else if (isNextResidue(Atoms[index], Atoms[i]))
			{
				if (Atoms[i].AtomName=="N")
				{
					SafePushBack(Atoms[index].connectivity, i, "connectivity");
				}
			}
		}
		else if (Atoms[index].AtomName=="O")
		{
			if (isSameResidue(Atoms[index], Atoms[i]))
			{
				if (Atoms[i].AtomName=="C")
				{
					SafePushBack(Atoms[index].connectivity, i, "connectivity");
				}
			}
		}
		else if (Atoms[index].AtomName=="N")
		{
			if (isPreviousResidue(Atoms[index], Atoms[i]))
			{
				if (Atoms[i].AtomName=="C")
				{
					SafePushBack(Atoms[index].connectivity, i, "connectivity");
				}
			}
			else if (isNextResidue(Atoms[index], Atoms[i]))
			{
				if (Atoms[i].AtomName=="CA")
				{
					SafePushBack(Atoms[index].connectivity, i, "connectivity");
				}
			}
		}
	}
}

void setConnectivity(vector<AtomStruct> &Atoms)
{
	int natom=Atoms.size();
	for (int i=0;i<natom;i++)
	{
		setConnectivity(Atoms, i);
	}
}

void setBondAnglesForAtom(ProteinStruct &Protein, int index)
{
	int nbond=Protein.Atoms[index].connectivity.size();
	BondAngleStruct tempBond;
	for (int i=0;i<nbond;i++)
	{
		for (int j=i+1;j<nbond;j++)
		{
			tempBond.atom1=Protein.Atoms[index].connectivity[i];
			tempBond.atom2=index;
			tempBond.atom3=Protein.Atoms[index].connectivity[j];
			SafePushBack(Protein.bondAngles, tempBond, "bondAngles");
		}
	}
}

void setBondAngles(ProteinStruct &Protein)
{
	int natom=Protein.Atoms.size();

	for (int i=0;i<natom;i++)
	{
		setBondAnglesForAtom(Protein, i);
	}
}

void setConnectivityAndBondAngles(ProteinStruct &Protein)
{
	setConnectivity(Protein.Atoms);
	setBondAngles(Protein);
}
